--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, TBOOX Open Source Group.
--
-- @author      ruki
-- @file        menu.lua
--

-- define module
local sandbox_core_project_menu = sandbox_core_project_menu or {}

-- load modules
local config    = require("project/config")
local project   = require("project/project")

-- get the menu options
function sandbox_core_project_menu.options()

    -- get it
    return project.menu()
end

-- return module
return sandbox_core_project_menu
